/*
 * ACB-BF592 W25Q16JV DPIAhCo
 *
 * Copyright(C) 2023 Kaneko System Co., Ltd
 *
 * Version : 0.0.1
 *
 * [History]
 *          Date       Author                     Version     Description
 * ---------------------------------------------------------------------------
 * CREATE : 01/Apr/2023 Shinya Kaneko             v.0.0.1     First release
 *
 * [Note]
 *
 */
#include <sys/platform.h>
#include <string.h>
#include "../common/flash.h"

/* STATUS bits */

#define STATUS1_BUSY        0x01
#define STATUS1_WEL         0x02

/* Various sizes */
#define PAGE_SIZE           0x100
#define SECTOR_SIZE         0x10000
#define NUM_SECTORS         32

#define BAUD_RATE_DIVISOR   2

static int w25q16jv_disable_quad_mode(const struct flash_info *fi);
static int w25q16jv_read_status(const struct flash_info *fi, uint8_t *status);

static void bf53x_spi_init(void)
{
    *pPORTF_FER   &= ~PF8;
    *pPORTFIO_SET  = PF8;
    *pPORTFIO_DIR |= PF8;
    /* SPI SCK, MOSI, MISO */
    *pPORTF_FER   |= PF13 | PF14 | PF15;
    *pPORTF_MUX   &= ~(PF13 | PF14 | PF15);

    *pSPI0_BAUD    = 2; /* 25MHz(MAX) */
    *pSPI0_CTL     = SPE | MSTR | EMISO | TDBR_CORE;
    *pSPI0_FLG     = 0;
    ssync();
    return;
}

static void bf53x_spi_close(void)
{
    return;
}

static void bf53x_spi_write(const uint8_t* p_tx_data, int32_t count)
{
    int32_t l_txc;
    volatile uint16_t l_dmy;

    for(l_txc = 0; l_txc < count; l_txc++)
    {
        *pSPI0_TDBR = p_tx_data[l_txc];
        ssync();
        while((*pSPI0_STAT & RXS) == 0) ssync();
        l_dmy = *pSPI0_RDBR;
    }
    return;
}

static void bf53x_spi_read(uint8_t* p_rx_data, int32_t count)
{
    int32_t l_rxc;

    for(l_rxc = 0; l_rxc < count; l_rxc++)
    {
        *pSPI0_TDBR = 0U;
        ssync();
        while((*pSPI0_STAT & RXS) == 0) ssync();
        p_rx_data[l_rxc] = (uint8_t)*pSPI0_RDBR;
    }
    return;
}

static void bf53x_spi_cs_assert(void)
{
    int32_t l_lc;

    *pPORTFIO_CLEAR = PF8;
    for(l_lc = 0; l_lc < 10; l_lc++) { ssync(); }
    return;
}

static void bf53x_spi_cs_deassert(void)
{
    int32_t l_lc;

    *pPORTFIO_SET = PF8;
    for(l_lc = 0; l_lc < 10; l_lc++) { ssync(); }
}


static int w25q16jv_open(struct flash_info *fi)
{
    uint8_t l_config;

    bf53x_spi_init();

    fi->size = NUM_SECTORS * SECTOR_SIZE;
    fi->number_of_regions = 1;
    fi->erase_block_regions = malloc(fi->number_of_regions * sizeof (struct erase_block_region));

    if (fi->erase_block_regions == NULL)
        return -1;

    fi->erase_block_regions[0].block_size = SECTOR_SIZE;
    fi->erase_block_regions[0].number_of_blocks = NUM_SECTORS;
    return 0;
}

static int w25q16jv_close(struct flash_info *fi)
{
    bf53x_spi_close();
    return 0;
}

static int w25q16jv_write_enable(const struct flash_info *fi)
{
    uint8_t l_cmd[1];
    uint8_t l_status;

    do
    {
        l_cmd[0] = 0x06U;
        bf53x_spi_cs_assert();
        bf53x_spi_write(&l_cmd[0], 1);
        bf53x_spi_cs_deassert();

        w25q16jv_read_status(fi, &l_status);
    } while((l_status & 0x02U) == 0U);

    return 0;
}

static int w25q16jv_write_disable(const struct flash_info *fi)
{
    uint8_t l_cmd[2];

    l_cmd[0] = 0x04U;
    bf53x_spi_cs_assert();
    bf53x_spi_write(&l_cmd[0], 1);
    bf53x_spi_cs_deassert();

    return 0;
}

static int w25q16jv_read_mid_did(struct flash_info *fi, uint8_t *mid, uint8_t *did)
{
    uint8_t l_cmd[4];

    /* Only for STANDARD mode */
    if (fi->current_mode != STANDARD)
        return -1;

    bf53x_spi_cs_assert();
    l_cmd[0] = 0x9FU;
    bf53x_spi_write(&l_cmd[0], 1);
    bf53x_spi_read(&l_cmd[0], 4);
    bf53x_spi_cs_deassert();

    *mid = l_cmd[0];
    *did = l_cmd[1];
    return 0;
}

static int w25q16jv_read_uid(const struct flash_info *fi, uint64_t *uid)
{
    /* Only for STANDARD mode */
    if (fi->current_mode != STANDARD)
        return -1;

    *uid = 0U;
    return 0;
}

static int w25q16jv_read_jedec_id(const struct flash_info *fi, uint8_t *mid,
        uint8_t *memory_type_id, uint8_t *capacity_id)
{
    uint8_t l_cmd[4];

    /* Only for STANDARD mode */
    if (fi->current_mode != STANDARD)
        return -1;

    /* :Single Modep */
    bf53x_spi_cs_assert();
    l_cmd[0] = 0x9E;
    bf53x_spi_write(&l_cmd[0], 1);
    bf53x_spi_read(&l_cmd[0], 3);
    bf53x_spi_cs_deassert();

    *mid = l_cmd[0];
    *memory_type_id = l_cmd[1];
    *capacity_id = l_cmd[2];
    return 0;
}

static int w25q16jv_read_status(const struct flash_info *fi, uint8_t *status)
{
    uint8_t l_cmd[2];

    l_cmd[0] = 0x05;
    bf53x_spi_cs_assert();
    bf53x_spi_write(&l_cmd[0], 1);
    bf53x_spi_read(&l_cmd[1], 1);
    bf53x_spi_cs_deassert();
    *status = l_cmd[1];
    return 0;
}

static int w25q16jv_not_busy(const uint8_t status)
{
    return (status & STATUS1_BUSY) ? 0 : 1;
}

static int w25q16jv_wait_ready(const struct flash_info *fi)
{
    uint8_t status;

    do
    {
        w25q16jv_read_status(fi, &status);
    } while(!w25q16jv_not_busy(status));

    return 0;
}

static int w25q16jv_write_status(const struct flash_info *fi, uint8_t status1, uint8_t status2)
{
    return 0;
}

static int w25q16jv_enable_quad_mode(const struct flash_info *fi)
{
    return 0;
}

static int w25q16jv_disable_quad_mode(const struct flash_info *fi)
{
    return 0;
}

static int w25q16jv_read(const struct flash_info *fi, uint32_t addr, uint8_t *buf, int size)
{
    uint8_t l_cmd[4];

    l_cmd[0] = 0x03;
    l_cmd[1] = (addr >> 16) & 0xFFU;
    l_cmd[2] = (addr >> 8) & 0xFFU;
    l_cmd[3] = (addr >> 0) & 0xFFU;
    bf53x_spi_cs_assert();
    bf53x_spi_write(l_cmd, 1+3);
    bf53x_spi_read(buf, size);
    bf53x_spi_cs_deassert();
    return 0;
}

static int w25q16jv_erase_1(const struct flash_info *fi,
        uint32_t addr, erase_type_t erase_type)
{
    uint8_t l_cmd[4];
    uint8_t l_flag;

    w25q16jv_write_enable(fi);

    if(erase_type == ERASE_CHIP)
    {
        l_cmd[0] = 0xC7;
        bf53x_spi_cs_assert();
        bf53x_spi_write(&l_cmd[0], 1);
        bf53x_spi_cs_deassert();
    }
    else
    {
        l_cmd[0] = 0xD8;
        l_cmd[1] = (addr >> 16) & 0xFFU;
        l_cmd[2] = (addr >> 8) & 0xFFU;
        l_cmd[3] = (addr >> 0) & 0xFFU;
        bf53x_spi_cs_assert();
        bf53x_spi_write(&l_cmd[0], 1+3);
        bf53x_spi_cs_deassert();
    }

    w25q16jv_wait_ready(fi);
    return 0;
}

static int w25q16jv_erase(const struct flash_info *fi, uint32_t addr, int size)
{
    size += addr % SECTOR_SIZE;
    addr -= addr % SECTOR_SIZE;

    while(size > 0)
    {
        w25q16jv_erase_1(fi, addr, ERASE_64KB);
        addr += SECTOR_SIZE;
        size -= SECTOR_SIZE;
    }

    return 0;
}

static int w25q16jv_erase_chip(const struct flash_info *fi)
{
    return w25q16jv_erase_1(fi, 0, ERASE_CHIP);
}

static int w25q16jv_page_program(struct flash_info *fi,
        uint32_t addr, const uint8_t *buf, int size)
{
    uint8_t l_cmd[4];
    uint8_t l_flag;

    w25q16jv_write_enable(fi);

    l_cmd[0] = 0x02;
    l_cmd[1] = (addr >> 16) & 0xFFU;
    l_cmd[2] = (addr >> 8) & 0xFFU;
    l_cmd[3] = (addr >> 0) & 0xFFU;
    bf53x_spi_cs_assert();
    bf53x_spi_write(&l_cmd[0], 4);
    bf53x_spi_write(buf, size);
    bf53x_spi_cs_deassert();

    w25q16jv_wait_ready(fi);
    return 0;
}

/* Program arbitrary size */
static int w25q16jv_program(struct flash_info *fi,
        uint32_t addr, const uint8_t *buf, int size)
{
    int program_size;

    while (size > 0)
    {
        if (addr % PAGE_SIZE != 0)
            program_size = ((PAGE_SIZE - addr % PAGE_SIZE) < size
                    ? (PAGE_SIZE - addr % PAGE_SIZE) : size);
        else if (size >= PAGE_SIZE)
            program_size = PAGE_SIZE;
        else
            program_size = size;

        w25q16jv_page_program(fi, addr, buf, program_size);

        addr += program_size;
        buf += program_size;
        size -= program_size;
    }

    return 0;
}

static int w25q16jv_reset(const struct flash_info *fi)
{
    /* Not supported */
    return -1;
}

struct flash_info w25q16jv_info =
{
    "W25Q16JV", /* name */
    "Winbond",  /* mname */
    0xEF,       /* manufacturer ID */
    0x40,       /* device ID */
    0x00,       /* memory type ID */
    0x15,       /* capacity ID */
    STANDARD,   /* supported modes */


    /* The following three fields are better be
       initialized in flash_open functions.  */
    0,          /* size in Byte */
    0,          /* number_of_regions */
    NULL,       /* erase_block_regions */
    127,        /* max_freq */
    127,        /* max_quad_freq */
    50,         /* max_read_data_freq */

    0,          /* cpha */
    0,          /* cpol */
    0,          /* lsb_first */
    0,          /* start_on_mosi */

    0,          /* frequency */
    STANDARD,   /* current_mode */

    w25q16jv_open,
    w25q16jv_close,

    w25q16jv_read_mid_did,
    w25q16jv_read_uid,
    w25q16jv_read_jedec_id,
    w25q16jv_read_status,
    w25q16jv_read_status,
    w25q16jv_write_status,
    w25q16jv_enable_quad_mode,
    w25q16jv_disable_quad_mode,
    w25q16jv_read,
    w25q16jv_write_enable,
    w25q16jv_write_disable,
    w25q16jv_erase,
    w25q16jv_erase_chip,
    w25q16jv_program,
    w25q16jv_reset,
};
